/**************************************************************************\
*   BC++ 1.01, GNU C++ 2.7.2                    Source: PWCOUNT.C
*   Host platform:   DOS, Linux                 Code:   ASCII-7
*   Target platform: DOS, *NIX                  Code:   ASCII-7
*   Project name:    CGI web-counter
*   Version:         0.92.00                    Started at:   06.08.97
*   Last change:     05.11.97                   Elapsed time: 0018 hours
*   Programmer:      Andrew Porokhnyak
****************************************************************************
*
* This source can be distributed and changed free. You may want to leave my
* copyright notices unchanged.
*
* Before you will compile this source, please check environment variables
* name (in preprocessor macro section) to match you httpd set up.
*
* Also you may change value of MAX_URLS (maximum HTML documents in a served
* tree) from 256 by default.
*
\**************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

/*** preprocessor macros **************************************************/

#define MAX_URLS        256             /* maximum documents counted */
#define OVERRIDE_PATH   ""              /* if blank string, then log and
        counters files will be placed in the same directory and will have
        the same name as CGI program file; else OVERRIDE_PATH+".cnt",
        +".log" will be used */

/* httpd-specific emvironmetn variables names */
#define ENV_DATE_LOC    "DATE_LOCAL"
#define ENV_DOC_URI     "DOCUMENT_URI"
#define ENV_SCRIPT_PATH "SCRIPT_FILENAME"
#define ENV_REMOTE_ADDR "REMOTE_ADDR"
#define ENV_REMOTE_HOST "REMOTE_HOST"
#define ENV_USER_AGENT  "HTTP_USER_AGENT"
#define ENV_USER_NAME   "USER_NAME"

/* error codes */
#define ERR_NO_PATH     11
#define ERR_NO_URI      12
#define ERR_NO_R_INFO   13

#define ERR_LOG_OPEN    21
#define ERR_LOG_WRITE   22
#define ERR_LOG_SEEK    23

#define ERR_CNT_OPEN    31
#define ERR_CNT_WRITE   32
#define ERR_CNT_ALLOC   33
#define ERR_CNT_OVER    34


/*** function prototypes **************************************************/

int init_variables(char *fcnt_name, char *flog_name, char *doc_uri,
    char *rem_addr, char *rem_host, char *usr_agnt, char *usr_name,
    char *cur_time);
int process_log(char *flog_name, char *rem_addr, char *rem_host,
    char *usr_agnt, char *doc_uri, char *cur_time);
int process_counter(char *fcnt_name, char *doc_uri, unsigned long *cnt);


/***************************************************************************
*   main()
*
* Called from HTML document by httpd. Perform a changes to counter file, log
* file and do CGI output of counter value
*
* Return: 0 everywhere
*
*       18.09.1997, Porokh
***************************************************************************/
int main(int argc, char **argv) {
    /* local vars */
    const char prod_name[]="PWCount";
    const char prod_ver1=0;
    const char prod_ver2=92;
    const char prod_ver3=0;
    const char prod_date[]="971205";
    const char prod_copr[]="(C) 1997 AOP";
    const char prod_help[]=
        "This is an CGI programm for count accesses to the Web pages\n"
        "It store all counters into one file pwcount.cnt in form:\n"
        "  <doc_uri>: <counter>\n"
        "Also it log access to HTML documents into a file pwcount.log:\n"
        "  <time> <user>@<host> (<addr>) <browser>\n"
        "To count your page, insert <!--#EXEC CGI=\"pwcount.cgi\" -->\n"
        ;

    char fcnt_name[256];        /* counter file name */
    char flog_name[256];        /* log file name */
    char doc_uri[256]="<none>"; /* current HTML doc URI */
    char rem_addr[16]="";       /* remote system IP addr */
    char rem_host[256]="";      /* remote system host name */
    char usr_name[64]="";       /* remote system user name */
    char usr_agnt[128]="";      /* remote system browser agent */
    char cur_time[20]="";       /* current time string */
    unsigned long cnt=0;        /* access counter */
    int ret_ini=0, ret_log=0, ret_cnt=0;    /* functions return codes */


    /* Check for arguments in cmd line (only "-?" now) */
    if ((argc==2) && strstr(argv[1],"-?")) {
        fprintf(stdout, "\n\t%s v%d.%02d.%02d (%s) %s\n\tCompiled: %s by %s\n",
            prod_name, prod_ver1, prod_ver2, prod_ver3, prod_date,
            prod_copr, __DATE__, "ANSI C");
        fprintf(stdout, "\n%s", prod_help);
        return 0;
    }

    /* Init values of the neccessary variables */
    ret_ini=init_variables(fcnt_name, flog_name, doc_uri, rem_addr,
        rem_host, usr_agnt, usr_name, cur_time);
    /* If no path info, stop execution */
    if (ret_ini == ERR_NO_PATH)
        return 0;
    /* Record remote system info into log file */
    ret_log=process_log(flog_name, rem_addr, rem_host, usr_agnt,
        doc_uri, cur_time);
    /* Increase counter value, corresponding to doc_uri */
    ret_cnt=process_counter(fcnt_name, doc_uri, &cnt);

    /* Perform actual CGI output */
    fprintf(stdout, "Content-type: text/html \n\n<HTML><BODY>\n");
    fprintf(stdout, "%ld\n", cnt);
    /* If an error occurs, write functions return codes */
    if (ret_ini || ret_log || ret_cnt)
        fprintf(stdout, "err(%i:%i:%i)\n", ret_ini, ret_log, ret_cnt);
    fprintf(stdout, "</BODY></HTML>\n");

    return 0;
} /*----------------------------------------------------------------------*/


/***************************************************************************
*   init_variables()
*
* Set up variables values from environment
*
* Return: 0 if all OK, !0 if something wrong
* Change: *fcnt_name, *flog_name, *doc_uri, *rem_addr, *rem_host,
*   *usr_agent, *usr_name, *cur_time
*
*       05.12.1997, Porokh
***************************************************************************/
int init_variables(
        char *fcnt_name,        /* counters file name */
        char *flog_name,        /* all logs file name */
        char *doc_uri,          /* current HTML document's URI */
        char *rem_addr,         /* remote system IP addr */
        char *rem_host,         /* remote system host name */
        char *usr_agnt,         /* remote system browser agent */
        char *usr_name,         /* remote system user name */
        char *cur_time          /* current time string */
        ) {

    char base_name[256];        /* base name w/o extension */
    time_t ti;                  /* current date&time (in seconds) */
    struct tm *tr;              /* date_time rec, global! */
    char *p;
    int retcode=0;

    /* Set ascii value of current date */
    tzset();
    time(&ti);
    tr=localtime(&ti);
    sprintf(cur_time,"%02d.%02d.%d %02d:%02d", tr->tm_mday,
        tr->tm_mon+1, tr->tm_year+1900, tr->tm_hour, tr->tm_min);
    /*fprintf(stdout, "%s\n", cur_time);*/

    /* Get values of the certain environment variables */
    if ( ((*(p=OVERRIDE_PATH)) != NULL) ||      /* if OVERRIDE_PATH not */
     ((p=getenv(ENV_SCRIPT_PATH)) != NULL) )    /* blank -- use it */
        strcpy(base_name, p);
    else
        return ERR_NO_PATH;     /* exit if no path info */
    if ((p=getenv(ENV_DOC_URI)) != NULL)
        strcpy(doc_uri, p);
    else
        retcode=ERR_NO_URI;
    if ((p=getenv(ENV_REMOTE_ADDR)) != NULL)
        strcpy(rem_addr, p);
    else
        retcode=ERR_NO_R_INFO;
    if ((p=getenv(ENV_REMOTE_HOST)) != NULL)
        strcpy(rem_host, p);
    else
        retcode=ERR_NO_R_INFO;
    if ((p=getenv(ENV_USER_AGENT)) != NULL)
        strcpy(usr_agnt, p);
    else
        retcode=ERR_NO_R_INFO;
    if ((p=getenv(ENV_USER_NAME)) != NULL)
        strcpy(usr_name, p);
    else
        retcode=ERR_NO_R_INFO;

    /* Remove last '.*' from base_name (if any) */
    for (p=base_name+strlen(base_name);p>base_name;p--) {
        if ((*p=='/') || (*p=='\\')) break;
        if (*p=='.') *p='\0';
    }

    strcpy(fcnt_name, base_name);
    strcat(fcnt_name, ".cnt");

    strcpy(flog_name, base_name);
    strcat(flog_name, ".log");

    return retcode;
} /*----------------------------------------------------------------------*/


/***************************************************************************
*   process_log()
*
* Write info about remote system into a log file
*
* Return: 0 if all OK, !0 if something wrong
* Change: none
*
*       18.09.1997, Porokh
***************************************************************************/
int process_log(
        char *flog_name,        /* log file name */
        char *rem_addr,         /* remote system IP addr */
        char *rem_host,         /* remote system host name */
        char *usr_agnt,         /* remote system browser agent */
        char *doc_uri,          /* requested document URI */
        char *cur_time          /* current time string */
        ) {

    FILE *flog;                 /* log file stream */

    if ((flog=fopen(flog_name, "at")) == NULL)
        return ERR_LOG_OPEN;
    if (fprintf(flog, "%s %s %s (%s) %s\n",
        cur_time, doc_uri, rem_host, rem_addr, usr_agnt)==EOF) {
        fclose(flog);
        return ERR_LOG_WRITE;
    }
    fclose(flog);

    return 0;
} /*----------------------------------------------------------------------*/


/***************************************************************************
*   process_counter()
*
* Read counter file <fcnt_name> line-by-line, store documents URI and
* corresponding counters into dynamic buffer. If given URI match <doc_uri>,
* then counter++
*
* Return: 0 if all OK, !0 if something wrong
* Change: *cnt
*
*       08.08.1997, Porokh
***************************************************************************/
int process_counter(
        char *fcnt_name,        /* counters file name */
        char *doc_uri,          /* current HTML document's URI */
        unsigned long *cnt      /* access counter for this document */
        ) {

    FILE *fcnt;                 /* counters file stream */
    char p[256], s[64];         /* temp string */
    char *buff[MAX_URLS];       /* buffer of lines */
    unsigned line=0;            /* line counts */
    char uri_found=0;           /* semaphore */
    unsigned i;
    int retcode=0;

    /* Try to read counters file (if exist) */
    if ((fcnt=fopen(fcnt_name, "rt")) != NULL) {
        while (1) {
            /* Break cycle, if EOF or invalid line */
            if (feof(fcnt)) break;
            if (fscanf(fcnt, "%s %s", p, s)<2) break;
            /* Check for URI matching, increment cnt if match */
            if (strstr(p, doc_uri)!=NULL) {
                uri_found=1;    /* touch-up semaphore */
                *cnt=atol(s);
                (*cnt)++;
                sprintf(s, "%li", *cnt);
            }
            /* Allocate space in buffer, exit if fails */
            if ((buff[line]=(char *)malloc(strlen(p)+strlen(s)+3))
                == NULL) {
                fclose(fcnt);
                return ERR_CNT_ALLOC;
            }
            /* Store URI and counter info in buffer */
            sprintf(buff[line++], "%s %s", p, s);
            /* Check for max url's */
            if (line >= MAX_URLS) {
                fclose(fcnt);
                return ERR_CNT_OVER;
            }
        }
    }

    /* Rewrite counters file, exit when fail */
    if ((fcnt=fopen(fcnt_name, "wt")) == NULL)
        return ERR_CNT_OPEN;
    /* Put stored URI and counter info back to the file */
    for (i=0; i<line; i++) {
        if (fprintf(fcnt, "%s\n", buff[i]) == EOF)
            retcode=ERR_CNT_WRITE;
        free(buff[i]);          /* free buffer element */
    }
    /* If current document's URI not found, append it */
    if (!uri_found) {
        *cnt=1;
        if (fprintf(fcnt, "%s: %li\n", doc_uri, *cnt) == EOF)
            retcode=ERR_CNT_WRITE;
    }
    fclose(fcnt);

    return retcode;
} /*----------------------------------------------------------------------*/


/* EOF PWCOUNT.C **********************************************************/
